'''subcommands/test.py'''

"""
Test auto-editor and make sure everything is working.
"""

# Internal Libraries
import os
import sys
import shutil
import platform
import subprocess
from time import perf_counter

# Included Libraries
from auto_editor.utils.func import clean_list
import auto_editor.vanparse as vanparse

def test_options(parser):
    parser.add_argument('--ffprobe_location', default='ffprobe',
        help='point to your custom ffprobe file.')
    parser.add_argument('--only', '-n', nargs='*')
    parser.add_argument('--help', '-h', action='store_true',
        help='print info about the program or an option and exit.')
    return parser

class FFprobe():
    def __init__(self, path):
        self.path = path

    def run(self, cmd):
        cmd.insert(0, self.path)
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)
        stdout, __ = process.communicate()
        return stdout.decode('utf-8')

    def pipe(self, cmd):
        full_cmd = [self.path, '-v', 'error'] + cmd
        process = subprocess.Popen(full_cmd, stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)
        stdout, __ = process.communicate()
        return stdout.decode('utf-8')

    def _get(self, file, stream, the_type, track, of='compact=p=0:nk=1'):
        return self.pipe(['-select_streams', '{}:{}'.format(the_type, track),
            '-show_entries', 'stream={}'.format(stream), '-of', of, file]).strip()

    def getResolution(self, file):
        return self._get(file, 'height,width', 'v', 0, of='csv=s=x:p=0')

    def getTimeBase(self, file):
        return self.pipe(['-select_streams', 'v', '-show_entries',
            'stream=avg_frame_rate', '-of', 'compact=p=0:nk=1', file]).strip()

    def getFrameRate(self, file):
        nums = clean_list(self.getTimeBase(file).split('/'), '\r\t\n')
        return int(nums[0]) / int(nums[1])

    def getAudioCodec(self, file, track=0):
        return self._get(file, 'codec_name', 'a', track)

    def getVideoCodec(self, file, track=0):
        return self._get(file, 'codec_name', 'v', track)

    def getSampleRate(self, file, track=0):
        return self._get(file, 'sample_rate', 'a', track)

    def getVLanguage(self, file):
        return self.pipe(['-show_entries', 'stream=index:stream_tags=language',
            '-select_streams', 'v', '-of', 'compact=p=0:nk=1', file]).strip()

    def getALanguage(self, file):
        return self.pipe(['-show_entries', 'stream=index:stream_tags=language',
            '-select_streams', 'a', '-of', 'compact=p=0:nk=1', file]).strip()

    def AudioBitRate(self, file):

        def bitrate_format(num):
            magnitude = 0
            while abs(num) >= 1000:
                magnitude += 1
                num /= 1000.0
            num = round(num)
            return '%d%s' % (num, ['', 'k', 'm', 'g', 't', 'p'][magnitude])

        exact_bitrate = self._get(file, 'bit_rate', 'a', 0)
        return bitrate_format(int(exact_bitrate))


def pipe_to_console(cmd):
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = process.communicate()
    return process.returncode, stdout.decode('utf-8'), stderr.decode('utf-8')

def cleanup(the_dir):
    for item in os.listdir(the_dir):
        item = os.path.join(the_dir, item)
        if('_ALTERED' in item or item.endswith('.xml') or item.endswith('.json')
            or item.endswith('.fcpxml') or item.endswith('.mlt')):
            os.remove(item)
        if(item.endswith('_tracks')):
            shutil.rmtree(item)

def clean_all():
    cleanup('resources')
    cleanup(os.getcwd())


def getRunner():
    if(platform.system() == 'Windows'):
        return ['py', '-m', 'auto_editor']
    return ['python3', '-m', 'auto_editor']


def run_program(cmd):
    no_open = '.' in cmd[0]
    cmd = getRunner() + cmd

    if(no_open):
        cmd += ['--no_open']

    returncode, stdout, stderr = pipe_to_console(cmd)
    if(returncode > 0):
        raise Exception('{}\n{}\n'.format(stdout, stderr))


def check_for_error(cmd, match=None):
    returncode, stdout, stderr = pipe_to_console(getRunner() + cmd)
    if(returncode > 0):
        if('Error!' in stderr):
            if(match is not None and match not in stderr):
                raise Exception('Could\'t find "{}"'.format(match))
        else:
            raise Exception('Program crashed.\n{}\n{}'.format(stdout, stderr))
    else:
        raise Exception('Program should not respond with a code 0.')

def inspect(path, *args):
    if(not os.path.exists(path)):
        raise Exception(f"Path '{path}' does not exist.")

    for item in args:
        func = item[0]
        expectedOutput = item[1]
        if(func(path) != expectedOutput):
            # Cheating on float numbers to allow 30 to equal 29.99944409236961
            if(isinstance(expectedOutput, float)):
                from math import ceil
                if(ceil(func(path) * 100) == expectedOutput * 100):
                    continue
            if(expectedOutput.endswith('k')):
                a = int(func(path)[:-1])
                b = int(expectedOutput[:-1])

                # Allow bitrate to have slight differences.
                if(abs(a - b) < 2):
                    continue

            raise Exception(
                f'Inspection Failed. Was {func(path)}, Expected {expectedOutput}.')


def make_np_list(in_file, compare_file, the_speed):
    import numpy as np
    from auto_editor.scipy.wavfile import read
    from auto_editor.audiotsm2 import phasevocoder
    from auto_editor.audiotsm2.io.array import ArrReader, ArrWriter

    samplerate, sped_chunk = read(in_file)
    spedup_audio = np.zeros((0, 2), dtype=np.int16)
    channels = 2

    with ArrReader(sped_chunk, channels, samplerate, 2) as reader:
        with ArrWriter(spedup_audio, channels, samplerate, 2) as writer:
            phasevocoder(reader.channels, speed=the_speed).run(
                reader, writer
            )
            spedup_audio = writer.output


    loaded = np.load(compare_file)

    if(not np.array_equal(spedup_audio, loaded['a'])):
        if(spedup_audio.shape == loaded['a'].shape):
            print(f'Both shapes ({spedup_audio.shape}) are same')
        else:
            print(spedup_audio.shape)
            print(loaded['a'].shape)

        result = np.subtract(spedup_audio, loaded['a'])

        print('result non-zero: {}'.format(np.count_nonzero(result)))
        print('len of spedup_audio: {}'.format(len(spedup_audio)))

        print(np.count_nonzero(result) / spedup_audio.shape[0], 'difference between arrays')

        raise Exception("file {} doesn't match array.".format(compare_file))

    # np.savez_compressed(out_file, a=spedup_audio)


class Tester():
    def __init__(self, args):
        self.passed_tests = 0
        self.failed_tests = 0
        self.args = args

    def run_test(self, name, func, description='', cleanup=None, allow_fail=False):
        if(self.args.only != [] and name not in self.args.only):
            return
        start = perf_counter()
        try:
            func()
            end = perf_counter() - start
        except Exception as e:
            self.failed_tests += 1
            print("Test '{}' failed.".format(name))
            print(e)
            if(not allow_fail):
                clean_all()
                sys.exit(1)
        else:
            self.passed_tests += 1
            print('{} Passed. {} secs.'.format(name, round(end, 2)))
            if(cleanup is not None):
                cleanup()

    def end(self):
        print('{}/{}'.format(self.passed_tests, self.passed_tests + self.failed_tests))
        clean_all()
        sys.exit(0)

def main(sys_args=None):
    parser = vanparse.ArgumentParser('test', 'version')
    parser = test_options(parser)

    if(sys_args is None):
        sys_args = sys.args[1:]

    try:
        args = parser.parse_args(sys_args)
    except vanparse.ParserError as e:
        print(e)
        sys.exit(1)

    ffprobe = FFprobe(args.ffprobe_location)

    tester = Tester(args)

    def help_tests():
        run_program(['--help'])
        run_program(['-h'])
        run_program(['--frame_margin', '--help'])
        run_program(['--frame_margin', '-h'])
        run_program(['--help', '--help'])
        run_program(['-h', '--help'])
        run_program(['--help', '-h'])
        run_program(['-h', '--help'])
    tester.run_test('help_tests', help_tests, description='check the help option, '
        'its short, and help on options and groups.')

    def version_debug():
        run_program(['--version'])
        run_program(['-v'])
        run_program(['-V'])

        run_program(['--debug'])

        # sanity check for example.mp4/ffprobe
        if(ffprobe.getFrameRate('example.mp4') != 30.0):
            print('getFrameRate did not equal 30.0')
            sys.exit(1)
    tester.run_test('version_tests', version_debug)


    def parser_test():
        check_for_error(['example.mp4', '--block'], 'needs argument')

    tester.run_test('parser_test', parser_test, allow_fail=True)

    def subtitle_tests():
        from auto_editor.render.subtitle import SubtitleParser
        test = SubtitleParser()
        test.contents = [
            [0, 10, "A"],
            [10, 20, "B"],
            [20, 30, "C"],
            [30, 40, "D"],
            [40, 50, "E"],
            [50, 60, "F"],
        ]
        speeds = [99999, 1]
        chunks = [[0, 10, 1], [10, 20, 0], [20, 30, 1], [30, 40, 0], [40, 50, 1],
            [50, 60, 0]]
        test.edit(chunks, speeds)

        if(test.contents != [[0, 10, "A"], [10, 20, "C"], [20, 30, "E"]]):
            raise ValueError('Incorrect subtitle results.')


    tester.run_test('subtitle_tests', subtitle_tests)

    def tsm_1a5_test():
        make_np_list('resources/example_cut_s16le.wav',
            'resources/example_1.5_speed.npz', 1.5)

    def tsm_0a5_test():
        make_np_list('resources/example_cut_s16le.wav',
            'resources/example_0.5_speed.npz', 0.5)

    def tsm_2a0_test():
        make_np_list('resources/example_cut_s16le.wav',
            'resources/example_2.0_speed.npz', 2)

    tester.run_test('tsm_1a5_test', tsm_1a5_test)
    tester.run_test('tsm_0a5_test', tsm_0a5_test, allow_fail=True)
    tester.run_test('tsm_2a0_test', tsm_2a0_test)

    def info_tests():
        run_program(['info', 'example.mp4'])
        run_program(['info', 'resources/man_on_green_screen.mp4'])
        run_program(['info', 'resources/multi-track.mov'])
        run_program(['info', 'resources/newCommentary.mp3'])
        run_program(['info', 'resources/test.mkv'])
    tester.run_test('info_tests', info_tests)

    def level_tests():
        run_program(['levels', 'resources/multi-track.mov'])
        run_program(['levels', 'resources/newCommentary.mp3'])
    tester.run_test('level_tests', level_tests)

    def example_tests():
        run_program(['example.mp4', '--has_vfr', 'no'])
        inspect(
            'example_ALTERED.mp4',
            [ffprobe.getFrameRate, 30.0],
            [ffprobe.getResolution, '1280x720'],
            [ffprobe.getVideoCodec, 'h264'],
            [ffprobe.getSampleRate, '48000'],
            [ffprobe.getVLanguage, '0|eng'],
            [ffprobe.getALanguage, '1|eng'],
        )
        run_program(['example.mp4', '--video_codec', 'uncompressed', '--has_vfr', 'no'])
        inspect(
            'example_ALTERED.mp4',
            [ffprobe.getFrameRate, 30.0],
            [ffprobe.getResolution, '1280x720'],
            [ffprobe.getVideoCodec, 'mpeg4'],
            [ffprobe.getSampleRate, '48000'],
        )
    tester.run_test('example_tests', example_tests)

    # Issue #200
    def url_test():
        run_program(['https://github.com/WyattBlue/auto-editor/raw/master/example.mp4'])
    tester.run_test('url_test', url_test)

    # Issue #172
    def bitrate_test():
        run_program(['example.mp4', '--audio_bitrate', '50k', '--has_vfr', 'no'])
        inspect(
            'example_ALTERED.mp4',
            [ffprobe.AudioBitRate, '50k'],
        )
    tester.run_test('bitrate_test', bitrate_test)

    # Issue #184
    def unit_tests():
        run_program(['example.mp4', '--mark_as_loud', '20s,22sec', '25secs,26.5seconds'])
        run_program(['example.mp4', '--sample_rate', '44100', '--has_vfr', 'no'])
        run_program(['example.mp4', '--sample_rate', '44100 Hz', '--has_vfr', 'no'])
        run_program(['example.mp4', '--sample_rate', '44.1 kHz', '--has_vfr', 'no'])
        run_program(['example.mp4', '--silent_threshold', '4%', '--has_vfr', 'no'])
    tester.run_test('unit_tests', unit_tests,
        description='''
        Make sure all units are working appropriately. That includes:
         - Seconds units: s, sec, secs, second, seconds
         - Frame units:   f, frame, frames
         - Sample units:  Hz, kHz
         - Percent:       %

        ''')

    def backwards_range_test():
        run_program(['example.mp4', '--edit', 'none', '--cut_out', '-5secs,end'])
        run_program(['example.mp4', '--edit', 'all', '--add_in', '-5secs,end'])
    tester.run_test('backwards_range_test', backwards_range_test, description='''
        Cut out the last 5 seconds of a media file by using negative number in the
        range.
        ''')

    def cut_out_test():
        run_program(['example.mp4', '--edit', 'none', '--video_speed', '2',
            '--silent_speed', '3', '--cut_out', '2secs,10secs', '--has_vfr', 'no'])
        run_program(['example.mp4', '--edit', 'all', '--video_speed', '2',
            '--add_in', '2secs,10secs', '--has_vfr', 'no'])
    tester.run_test('cut_out_test', cut_out_test)

    def gif_test():
        run_program(['resources/man_on_green_screen.gif', '--edit', 'none'])
        inspect(
            'resources/man_on_green_screen_ALTERED.gif',
            [ffprobe.getVideoCodec, 'gif'],
        )
    tester.run_test('gif_test', gif_test, description='''
        Feed auto-editor a gif file and make sure it can spit out a correctly formated
        gif. No editing is requested.
        ''',
        cleanup=clean_all)

    def margin_tests():
        run_program(['example.mp4', '-m', '3'])
        run_program(['example.mp4', '--margin', '3'])
        run_program(['example.mp4', '-m', '0.3sec'])
        run_program(['example.mp4', '-m', '6f'])
        run_program(['example.mp4', '-m', '5 frames'])
        run_program(['example.mp4', '-m', '0.4 seconds'])
    tester.run_test('margin_tests', margin_tests)

    def input_extension():
        # Input file must have an extension. Throw error if none is given.
        shutil.copy('example.mp4', 'example')
        check_for_error(['example', '--no_open'], 'must have an extension.')
        os.remove('example')

    tester.run_test('input_extension', input_extension)

    def output_extension():
        # Add input extension to output name if no output extension is given.
        run_program(['example.mp4', '--has_vfr', 'no', '-o', 'out'])
        inspect(
            'out.mp4',
            [ffprobe.getVideoCodec, 'h264']
        )
        os.remove('out.mp4')

        run_program(['resources/test.mkv', '--has_vfr', 'no', '-o', 'out'])
        inspect(
            'out.mkv',
            [ffprobe.getVideoCodec, 'h264']
        )
        os.remove('out.mkv')

    tester.run_test('output_extension', output_extension)

    def progress_ops_test():
        run_program(['example.mp4', '--has_vfr', 'no', '--machine_readable_progress'])
        run_program(['example.mp4', '--has_vfr', 'no', '--no_progress'])
    tester.run_test('progress_ops_test', progress_ops_test)

    def silent_threshold():
        run_program(['resources/newCommentary.mp3', '--silent_threshold', '0.1'])
    tester.run_test('silent_threshold', silent_threshold)

    def track_tests():
        run_program(['resources/multi-track.mov', '--cut_by_all_tracks'])
        run_program(['resources/multi-track.mov', '--keep_tracks_seperate'])
        run_program(['example.mp4', '--cut_by_this_audio', 'resources/newCommentary.mp3'])
    tester.run_test('track_tests', track_tests)

    def json_tests():
        run_program(['example.mp4', '--export_as_json'])
        run_program(['example.json'])
    tester.run_test('json_tests', json_tests)

    def scale_tests():
        run_program(['example.mp4', '--scale', '1.5', '--has_vfr', 'no'])
        inspect(
            'example_ALTERED.mp4',
            [ffprobe.getFrameRate, 30.0],
            [ffprobe.getResolution, '1920x1080'],
            [ffprobe.getSampleRate, '48000'],
        )

        run_program(['example.mp4', '--scale', '0.2', '--has_vfr', 'no'])
        inspect(
            'example_ALTERED.mp4',
            [ffprobe.getFrameRate, 30.0],
            [ffprobe.getResolution, '256x144'],
            [ffprobe.getSampleRate, '48000'],
        )
    tester.run_test('scale_tests', scale_tests)

    def various_errors_test():
        check_for_error(['example.mp4', '--add_rectangle', '0,60', '--cut_out', '60,end'])

    tester.run_test('various_errors_test', various_errors_test)

    def effect_tests():
        run_program(['create', 'test', '--width', '640', '--height', '360', '-o',
            'testsrc.mp4'])
        inspect(
            'testsrc.mp4',
            [ffprobe.getFrameRate, 30.0],
            [ffprobe.getResolution, '640x360'],
        )
        run_program(['testsrc.mp4', '--mark_as_loud', 'start,end', '--add_rectangle',
            '0,30,0,200,100,300,#43FA56,10'])
        os.remove('testsrc_ALTERED.mp4')
        os.remove('testsrc.mp4')
    tester.run_test('effect_tests', effect_tests,
        description='test the rectangle object',
        cleanup=clean_all)

    def render_text():
        run_program(['example.mp4', '--has-vfr', 'no', '--add-text', 'start,end,This is my text'])

    tester.run_test('render_text', render_text)

    def check_font_error():
        check_for_error(
            ['example.mp4', '--has-vfr', 'no', '--add-text', 'start,end,text,0,0,30,notafont'],
            'not found')

    tester.run_test('check_font_error', check_font_error)

    def export_tests():
        for item in os.listdir('resources'):
            if('man_on_green_screen' in item or item.startswith('.') or '_ALTERED' in item
                or item.endswith('.npz')):
                continue
            item = 'resources/{}'.format(item)
            run_program([item])
            run_program([item, '-exp'])
            run_program([item, '-exf'])
            run_program([item, '-exs'])
            run_program([item, '--export_as_clip_sequence'])
            run_program([item, '--preview'])
            cleanup('resources')
    tester.run_test('export_tests', export_tests)

    def codec_tests():
        run_program(['example.mp4', '--video_codec', 'h264', '--preset', 'faster'])
        run_program(['example.mp4', '--audio_codec', 'ac3'])
    tester.run_test('codec_tests', codec_tests)

    def combine_tests():
        run_program(['example.mp4', '--mark_as_silent', '0,171', '-o', 'hmm.mp4'])
        run_program(['example.mp4', 'hmm.mp4', '--combine_files', '--debug'])
        os.remove('hmm.mp4')
    tester.run_test('combine_tests', combine_tests)

    def motion_tests():
        run_program(['resources/man_on_green_screen.mp4', '--edit_based_on', 'motion',
         '--debug', '--frame_margin', '0', '-mcut', '0', '-mclip', '0'])
        run_program(['resources/man_on_green_screen.mp4', '--edit_based_on', 'motion',
            '--motion_threshold', '0'])
    tester.run_test('motion_tests', motion_tests)

    tester.end()

if(__name__ == '__main__'):
    main()



'''audiotsm2/wsola.py'''

import numpy as np

from auto_editor.audiotsm2.base import AnalysisSynthesisTSM
from auto_editor.audiotsm2.utils.windows import hanning

class WSOLAConverter():
    """
    A Converter implementing the WSOLA (Waveform Similarity-based Overlap-Add)
    time-scale modification procedure.
    """
    def __init__(self, channels, frame_length, synthesis_hop, tolerance):
        self._channels = channels
        self._frame_length = frame_length
        self._synthesis_hop = synthesis_hop
        self._tolerance = tolerance

        self._synthesis_frame = np.empty((channels, frame_length))
        self._natural_progression = np.empty((channels, frame_length))
        self._first = True

    def clear(self):
        self._first = True

    def convert_frame(self, analysis_frame):
        for k in range(0, self._channels):
            if self._first:
                delta = 0
            else:
                cross_correlation = np.correlate(
                    analysis_frame[k, :-self._synthesis_hop],
                    self._natural_progression[k])
                delta = np.argmax(cross_correlation)
                del cross_correlation

            # Copy the shifted analysis frame to the synthesis frame buffer
            np.copyto(self._synthesis_frame[k],
                      analysis_frame[k, delta:delta + self._frame_length])

            # Save the natural progression (what the next synthesis frame would
            # be at normal speed)
            delta += self._synthesis_hop
            np.copyto(self._natural_progression[k],
                      analysis_frame[k, delta:delta + self._frame_length])

        self._first = False

        return self._synthesis_frame


def wsola(channels, speed=1., frame_length=1024, analysis_hop=None, synthesis_hop=None,
    tolerance=None):

    if synthesis_hop is None:
        synthesis_hop = frame_length // 2

    if analysis_hop is None:
        analysis_hop = int(synthesis_hop * speed)

    if tolerance is None:
        tolerance = frame_length // 2

    analysis_window = None
    synthesis_window = hanning(frame_length)

    converter = WSOLAConverter(channels, frame_length, synthesis_hop, tolerance)

    return AnalysisSynthesisTSM(
        converter, channels, frame_length, analysis_hop, synthesis_hop,
        analysis_window, synthesis_window, tolerance,
        tolerance + synthesis_hop)



'''cutting.py'''

import numpy as np

from auto_editor.utils.types import split_num_str

def combine_audio_motion(audioList, motionList, based, log):
    # type: (np.ndarray, np.ndarray, str, Any) -> np.ndarray

    if(based == 'audio' or based == 'not_audio'):
        if(max(audioList) == 0):
            log.error('There was no place where audio exceeded the threshold.')
    if(based == 'motion' or based == 'not_motion'):
        if(max(motionList) == 0):
            log.error('There was no place where motion exceeded the threshold.')

    if(audioList is not None and max(audioList) == 0):
        log.warning('There was no place where audio exceeded the threshold.')
    if(motionList is not None and max(motionList) == 0):
        log.warning('There was no place where motion exceeded the threshold.')

    if(based == 'audio'):
        return audioList

    if(based == 'motion'):
        return motionList

    if(based == 'not_audio'):
        return np.invert(audioList)

    if(based == 'not_motion'):
        return np.invert(motionList)

    if(based == 'audio_and_motion'):
        return audioList & motionList

    if(based == 'audio_or_motion'):
        return audioList | motionList

    if(based == 'audio_xor_motion'):
        return np.bitwise_xor(audioList, motionList)

    if(based == 'audio_and_not_motion'):
        return audioList & np.invert(motionList)

    if(based == 'not_audio_and_motion'):
        return np.invert(audioList) & motionList

    if(based == 'not_audio_and_not_motion'):
        return np.invert(audioList) & np.invert(motionList)
    return None


def combine_segment(has_loud, segment, fps):
    for item in segment:
        start, end = item['segment']
        start = int(start * fps)
        end = int(end * fps)
        has_loud[start:end] = False
    return has_loud


def removeSmall(has_loud, lim, replace, with_):
    # type: (np.ndarray, int, int, int) -> np.ndarray
    startP = 0
    active = False
    for j, item in enumerate(has_loud):
        if(item == replace):
            if(not active):
                startP = j
                active = True
            # Special case for end.
            if(j == len(has_loud) - 1):
                if(j - startP < lim):
                    has_loud[startP:j+1] = with_
        else:
            if(active):
                if(j - startP < lim):
                    has_loud[startP:j] = with_
                active = False
    return has_loud


def str_is_number(val):
    # type: (str) -> bool
    return val.replace('.', '', 1).replace('-', '', 1).isdigit()


def str_starts_with_number(val):
    # type: (str) -> bool
    if(val.startswith('-')):
        val = val[1:]
    val = val.replace('.', '', 1)
    return val[0].isdigit()


def setRange(has_loud, range_syntax, fps, with_, log):
    # type: (...) -> np.ndarray

    def replace_variables_to_values(item, fps, log):
        # type: (str, float | int, Any) -> int
        if(str_is_number(item)):
            return int(item)
        if(str_starts_with_number(item)):
            value, unit = split_num_str(item, log.error)
            if(unit in ['', 'f', 'frame', 'frames']):
                if(isinstance(value, float)):
                    log.error('float type cannot be used with frame unit')
                return value
            if(unit in ['s', 'sec', 'secs', 'second', 'seconds']):
                return round(value * fps)
            log.error('Unknown unit: {}'.format(unit))
        if(item == 'start'):
            return 0
        if(item == 'end'):
            return len(has_loud)
        log.error("variable '{}' not available.".format(item))

    def var_val_to_frames(val, fps, log):
        # type: (str, float | int, Any) -> int
        num = replace_variables_to_values(val, fps, log)
        if(num < 0):
            num += len(has_loud)
        return num

    for item in range_syntax:
        pair = []
        for val in item:
            pair.append(var_val_to_frames(val, fps, log))
        has_loud[pair[0]:pair[1]] = with_
    return has_loud


def seconds_to_frames(value, fps):
    # type: (int | str, float) -> int
    if(isinstance(value, str)):
        return int(float(value) * fps)
    return value

def cook(has_loud, minClip, minCut):
    # type: (np.ndarray, int, int) -> np.ndarray
    has_loud = removeSmall(has_loud, minClip, replace=1, with_=0)
    has_loud = removeSmall(has_loud, minCut, replace=0, with_=1)
    return has_loud


# Turn long silent/loud array to formatted chunk list.
# Example: [1, 1, 1, 0, 0] => [[0, 3, 1], [3, 5, 0]]
def chunkify(has_loud, has_loud_length=None):
    # type: (np.ndarray, int | None) -> list[list[int]]
    if(has_loud_length is None):
        has_loud_length = len(has_loud)

    chunks = []
    startP = 0
    for j in range(1, has_loud_length):
        if(has_loud[j] != has_loud[j - 1]):
            chunks.append([startP, j, int(has_loud[j-1])])
            startP = j
    chunks.append([startP, has_loud_length, int(has_loud[j])])
    return chunks


# Turn chunk list into silent/loud like array.
# Example: [[0, 3, 1], [3, 5, 0]] => [1, 1, 1, 0, 0]
def chunks_to_has_loud(chunks):
    # type: (list[list[int]]) -> np.ndarray
    duration = chunks[len(chunks) - 1][1]
    has_loud = np.zeros((duration), dtype=np.uint8)

    for chunk in chunks:
        if(chunk[2] != 0):
            has_loud[chunk[0]:chunk[1]] = chunk[2]
    return has_loud


def apply_frame_margin(has_loud, has_loud_length, frame_margin):
    # type: (np.ndarray, int, int) -> np.ndarray
    new = np.zeros((has_loud_length), dtype=np.uint8)
    for i in range(has_loud_length):
        start = int(max(0, i - frame_margin))
        end = int(min(has_loud_length, i+1+frame_margin))
        new[i] = min(1, np.max(has_loud[start:end]))
    return new


def apply_mark_as(has_loud, has_loud_length, fps, args, log):
    # type: (...) -> np.ndarray
    if(args.mark_as_loud != []):
        has_loud = setRange(has_loud, args.mark_as_loud, fps, 1, log)

    if(args.mark_as_silent != []):
        has_loud = setRange(has_loud, args.mark_as_silent, fps, 0, log)
    return has_loud

def apply_spacing_rules(has_loud, has_loud_length, minClip, minCut, speeds, fps, args, log):
    # type: (...) -> list[list[int]]
    if(args.cut_out != []):
        has_loud = setRange(has_loud, args.cut_out, fps, speeds.index(99999), log)

    if(args.add_in != []):
        has_loud = setRange(has_loud, args.add_in, fps, speeds.index(args.video_speed), log)

    if(args.set_speed_for_range != []):
        for item in args.set_speed_for_range:
            my_speed_index = speeds.index(float(item[0]))
            has_loud = setRange(has_loud, [item[1:]], fps, my_speed_index, log)

    return chunkify(has_loud, has_loud_length)


def apply_basic_spacing(has_loud, fps, minClip, minCut, log):
    # type: (np.ndarray, float, int, int, Any) -> list[list[int]]
    minClip = seconds_to_frames(minClip, fps)
    minCut = seconds_to_frames(minCut, fps)

    has_loud = cook(has_loud, minClip, minCut)
    return chunkify(has_loud)


def merge(start_list, end_list):
    # type: (np.ndarray, np.ndarray) -> np.ndarray
    merge = np.zeros((len(start_list)), dtype=np.bool_)

    startP = 0
    for item in start_list:
        if(item == True):
            where_list = np.where(end_list[startP:])[0]
            if(len(where_list) > 0):
                merge[startP:where_list[0]] = True
        startP += 1
    return merge




'''render/audio.py'''

from auto_editor.scipy.wavfile import read

from auto_editor.audiotsm2 import phasevocoder
from auto_editor.audiotsm2.io.array import ArrReader, ArrWriter
from auto_editor.audiotsm2.io.wav import WavWriter

import numpy as np

def make_new_audio(input_path, output_path, chunks, speeds, log, fps, progress):

    if(len(chunks) == 1 and chunks[0][2] == 0):
        log.error('Trying to create an empty file.')

    samplerate, audio_samples = read(input_path)

    channels = 2
    samplewidth = 2

    progress.start(len(chunks), 'Creating new audio')

    with WavWriter(output_path, 2, samplerate) as main_writer:
        for c, chunk in enumerate(chunks):
            sample_start = int(chunk[0] / fps * samplerate)
            sample_end = int(sample_start + (samplerate / fps) * (chunk[1] - chunk[0]))

            the_speed = speeds[chunk[2]]

            if(the_speed == 1):
                main_writer.write(audio_samples[sample_start:sample_end].T / 32676)
            elif(the_speed != 99999):
                sped_chunk = audio_samples[sample_start:sample_end]
                spedup_audio = np.zeros((0, 2), dtype=np.int16)
                with ArrReader(sped_chunk, channels, samplerate, samplewidth) as reader:
                    with ArrWriter(spedup_audio, channels, samplerate, samplewidth) as writer:
                        phasevocoder(reader.channels, speed=the_speed).run(
                            reader, writer
                        )
                        spedup_audio = writer.output
                        main_writer.write(spedup_audio.T / 32676)

            progress.tick(c)
        progress.end()



'''utils/video.py'''

# Internal Libraries
import os.path

# Included Libraries
from .func import fnone

def fset(cmd, option, value):
    if(fnone(value)):
        return cmd
    return cmd + [option] + [value]


def get_vcodec(args, inp, rules):
    vcodec = args.video_codec
    if(vcodec == 'auto'):
        vcodec = inp.video_streams[0]['codec']

        if((rules['vstrict'] and vcodec not in rules['vcodecs'])
            or (vcodec in rules['disallow_v'])):
            return rules['vcodecs'][0]

    if(vcodec == 'copy'):
        return inp.video_streams[0]['codec']

    if(vcodec == 'uncompressed'):
        return 'mpeg4'
    return vcodec

def get_acodec(args, inp, rules):
    acodec = args.audio_codec
    if(acodec == 'auto'):
        acodec = inp.audio_streams[0]['codec']

        if((rules['astrict'] and acodec not in rules['acodecs'])
            or (acodec in rules['disallow_a'])):
            return rules['acodecs'][0]

    if(acodec == 'copy'):
        return inp.audio_streams[0]['codec']
    return acodec

def video_quality(cmd, args, inp, rules):
    cmd = fset(cmd, '-crf', args.constant_rate_factor)
    cmd = fset(cmd, '-b:v', args.video_bitrate)
    cmd = fset(cmd, '-tune', args.tune)
    cmd = fset(cmd, '-preset', args.preset)

    qscale = args.video_quality_scale

    if(args.video_codec == 'uncompressed' and fnone(qscale)):
        qscale = '1'

    vcodec = get_vcodec(args, inp, rules)

    cmd.extend(['-c:v', vcodec])

    cmd = fset(cmd, '-qscale:v', qscale)

    cmd.extend(['-movflags', 'faststart'])
    return cmd


def mux_quality_media(ffmpeg, video_stuff, rules, write_file, container, args, inp,
    temp, log):
    s_tracks = 0 if not rules['allow_subtitle'] else len(inp.subtitle_streams)
    a_tracks = 0 if not rules['allow_audio'] else len(inp.audio_streams)
    v_tracks = 0
    cmd = ['-hide_banner', '-y', '-i', inp.path]

    for _, spedup, _ in video_stuff:
        if(spedup is not None):
            cmd.extend(['-i', spedup])
            v_tracks += 1

    if(a_tracks > 0):
        if(args.keep_tracks_seperate and rules['max_audio_streams'] is None):
            for t in range(a_tracks):
                cmd.extend(['-i', os.path.join(temp, f'new{t}.wav')])
        else:
            # Merge all the audio a_tracks into one.
            new_a_file = os.path.join(temp, 'new_audio.wav')
            if(a_tracks > 1):
                new_cmd = []
                for t in range(a_tracks):
                    new_cmd.extend(['-i', os.path.join(temp, f'new{t}.wav')])
                new_cmd.extend(['-filter_complex', f'amerge=inputs={a_tracks}', '-ac',
                    '2', new_a_file])
                ffmpeg.run(new_cmd)
                a_tracks = 1
            else:
                new_a_file = os.path.join(temp, 'new0.wav')
            cmd.extend(['-i', new_a_file])

    if(s_tracks > 0):
        for s, sub in enumerate(inp.subtitle_streams):
            new_path = os.path.join(temp, 'new{}s.{}'.format(s, sub['ext']))
            cmd.extend(['-i', new_path])

    total_streams = v_tracks + s_tracks + a_tracks

    for i in range(total_streams):
        cmd.extend(['-map', f'{i+1}:0'])

    # Copy lang metadata
    for i, track in enumerate(inp.video_streams):
        if(i > v_tracks):
            break
        if(track['lang'] is not None):
            cmd.extend([f'-metadata:s:v:{i}', f'language={track["lang"]}'])

    for i, track in enumerate(inp.audio_streams):
        if(i > a_tracks):
            break
        if(track['lang'] is not None):
            cmd.extend([f'-metadata:s:a:{i}', f'language={track["lang"]}'])

    for i, track in enumerate(inp.subtitle_streams):
        if(i > s_tracks):
            break
        if(track['lang'] is not None):
            cmd.extend([f'-metadata:s:s:{i}', f'language={track["lang"]}'])

    for video_type, _, apply_video in video_stuff:
        if(video_type == 'video'):
            if(apply_video):
                cmd = video_quality(cmd, args, inp, rules)
            else:
                cmd.extend(['-c:v', 'copy'])
            break

    if(s_tracks > 0):
        scodec = inp.subtitle_streams[0]['codec']
        if(inp.ext == '.' + container):
            cmd.extend(['-c:s', scodec])
        elif(rules['scodecs'] is not None):
            if(scodec not in rules['scodecs']):
                scodec = rules['scodecs'][0]
            cmd.extend(['-c:s', scodec])

    if(a_tracks > 0):
        acodec = get_acodec(args, inp, rules)

        cmd = fset(cmd, '-c:a', acodec)
        cmd = fset(cmd, '-b:a', args.audio_bitrate)

        if(fnone(args.sample_rate)):
            if(rules['samplerate'] is not None):
                cmd.extend(['-ar', str(rules['samplerate'][0])])
        else:
            cmd.extend(['-ar', str(args.sample_rate)])

    cmd.extend(['-strict', '-2']) # Allow experimental codecs.
    cmd.extend(['-map', '0:t?', '-map', '0:d?']) # Add input attachments and data to output.
    cmd.append(write_file)
    ffmpeg.run_check_errors(cmd, log)



'''analyze/audio.py'''

import numpy as np
import math

def get_max_volume(s):
    # type: (np.ndarray) -> float
    return max(float(np.max(s)), -float(np.min(s)))


def display_audio_levels(read_track, fps):
    import sys

    from auto_editor.scipy.wavfile import read

    sample_rate, audio_samples = read(read_track)

    max_volume = get_max_volume(audio_samples)
    sample_count = audio_samples.shape[0]

    sample_rate_per_frame = sample_rate / fps
    audio_frame_count = math.ceil(sample_count / sample_rate_per_frame)

    for i in range(audio_frame_count):
        start = int(i * sample_rate_per_frame)
        end = min(int((i+1) * sample_rate_per_frame), sample_count)
        audiochunks = audio_samples[start:end]
        sys.stdout.write('{:.20f}\n'.format(get_max_volume(audiochunks) / max_volume))


def audio_detection(audio_samples, sample_rate, silent_threshold, fps, log):
    # type: (np.ndarray, int, float, float, Any) -> np.ndarray
    log.conwrite('Analyzing audio volume.')

    max_volume = get_max_volume(audio_samples)
    sample_count = audio_samples.shape[0]

    sample_rate_per_frame = sample_rate / fps
    audio_frame_count = math.ceil(sample_count / sample_rate_per_frame)
    has_loud_audio = np.zeros((audio_frame_count), dtype=np.bool_)

    # Calculate when the audio is loud or silent.
    for i in range(audio_frame_count):
        start = int(i * sample_rate_per_frame)
        end = min(int((i+1) * sample_rate_per_frame), sample_count)
        audiochunks = audio_samples[start:end]
        if(get_max_volume(audiochunks) / max_volume > silent_threshold):
            has_loud_audio[i] = True

    return has_loud_audio



'''subcommands/grep.py'''

import sys
import os
import re
import tempfile

def grep_options(parser):
    parser.add_argument('--no_filename', action='store_true',
        help='Never print filenames with output lines.')
    parser.add_argument('--max_count', '-m', type=int, default=float('inf'),
        help='Stop reading a file after NUM matching lines.')
    parser.add_argument('--count', '-c', action='store_true',
        help='Suppress normal output; instead print count of matching lines for each file.')
    parser.add_argument('--ignore_case', '-i', action='store_true',
        help='Ignore case distinctions for the PATTERN.')
    parser.add_argument('--timecode', action='store_true',
        help="Print the match's timecode.")
    parser.add_argument('--time', action='store_true',
        help="Print when the match happens. (Ignore ending).")
    parser.add_argument('--ffmpeg_location', default=None,
        help='Point to your custom ffmpeg file.')
    parser.add_argument('--my_ffmpeg', action='store_true',
        help='Use the ffmpeg on your PATH instead of the one packaged.')
    parser.add_argument('--help', '-h', action='store_true',
        help='Print info about the program or an option and exit.')
    parser.add_argument('input', nargs='*',
        help='The path to a file you want inspected.')
    return parser

# stackoverflow.com/questions/9662346/python-code-to-remove-html-tags-from-a-string
def cleanhtml(raw_html):
    # type: (str) -> str
    cleanr = re.compile('<.*?>')
    cleantext = re.sub(cleanr, '', raw_html)
    return cleantext

def grep_core(media_file, add_prefix, ffmpeg, args, log, TEMP):

    """
    We're using the WEBVTT subtitle format. It's better than srt
    because it doesn't emit line numbers and the time code is in
    (hh:mm:ss.sss) instead of (dd:hh:mm:ss,sss)
    """

    out_file = os.path.join(TEMP, 'media.vtt')
    ffmpeg.run(['-i', media_file, out_file])

    count = 0

    flags = 0
    if(args.ignore_case):
        flags = re.IGNORECASE

    prefix = ''
    if(add_prefix):
        prefix = '{}:'.format(os.path.splitext(os.path.basename(media_file))[0])

    timecode = ''
    line_number = -1
    with open(out_file, 'r') as file:
        while True:
            line = file.readline()
            line_number += 1
            if(line_number == 0):
                continue

            if(not line or count >= args.max_count):
                break

            if(line.strip() == ''):
                continue

            if(re.match(r'\d*:\d\d.\d*\s-->\s\d*:\d\d.\d*', line)):
                if(args.time):
                    timecode = line.split('-->')[0].strip() + ' '
                else:
                    timecode = line.strip() + '; '
                continue

            line = cleanhtml(line)
            match = re.search(args.input[0], line, flags)
            line = line.strip()

            if(match):
                count += 1
                if(not args.count):
                    if(args.timecode or args.time):
                        print(prefix + timecode + line)
                    else:
                        print(prefix + line)

    if(args.count):
        print(prefix + str(count))


def main(sys_args=sys.argv[1:]):
    import auto_editor
    import auto_editor.vanparse as vanparse

    from auto_editor.utils.log import Log
    from auto_editor.ffwrapper import FFmpeg

    parser = vanparse.ArgumentParser('grep', auto_editor.version,
        description='Read and match subtitle tracks in media files.')
    parser = grep_options(parser)

    TEMP = tempfile.mkdtemp()
    log = Log(temp=TEMP)

    try:
        args = parser.parse_args(sys_args)
    except vanparse.ParserError as e:
        log.error(str(e))

    ffmpeg = FFmpeg(args.ffmpeg_location, args.my_ffmpeg, debug=False)

    media_files = args.input[1:]

    add_prefix = (len(media_files) > 1 or os.path.isdir(media_files[0])) and not args.no_filename

    for media_file in media_files:
        if(not os.path.exists(media_file)):
            log.error('{}: File does not exist.'.format(media_file))

        if(os.path.isdir(media_file)):
            for _, _, files in os.walk(media_file):
                for file in files:
                    if(file == '.DS_Store'):
                        continue
                    grep_core(os.path.join(media_file, file),
                        add_prefix, ffmpeg, args, log, TEMP)
        else:
            grep_core(media_file, add_prefix, ffmpeg, args, log, TEMP)

    log.cleanup()

if(__name__ == '__main__'):
    main()



'''edit.py'''

# Internal Libraries
import os

import auto_editor
from auto_editor.utils.effects import Effect
from auto_editor.utils.func import fnone, set_output_name, append_filename

def get_chunks(inp, speeds, segment, fps, args, log, audio_samples=None, sample_rate=None):
    from auto_editor.cutting import (combine_audio_motion, combine_segment,
        apply_spacing_rules, apply_mark_as, apply_frame_margin, seconds_to_frames, cook)

    frame_margin = seconds_to_frames(args.frame_margin, fps)
    min_clip = seconds_to_frames(args.min_clip_length, fps)
    min_cut = seconds_to_frames(args.min_cut_length, fps)

    def get_has_loud(inp, args, fps, audio_samples, sample_rate, log):
        import numpy as np

        from auto_editor.analyze.generic import get_np_list

        if(args.edit_based_on == 'none'):
            return get_np_list(inp, audio_samples, sample_rate, fps, np.ones)
        if(args.edit_based_on == 'all'):
            return get_np_list(inp, audio_samples, sample_rate, fps, np.zeros)

        audio_list, motion_list = None, None

        if('audio' in args.edit_based_on):
            from auto_editor.analyze.audio import audio_detection

            if(audio_samples is None):
                audio_list = get_np_list(inp, audio_samples, sample_rate, fps, np.ones)
            else:
                audio_list = audio_detection(audio_samples, sample_rate,
                    args.silent_threshold, fps, log)

        if('motion' in args.edit_based_on):
            from auto_editor.analyze.motion import motion_detection

            if(len(inp.video_streams) == 0):
                motion_list = get_np_list(inp, audio_samples, sample_rate, fps, np.ones)
            else:
                motion_list = motion_detection(inp, args.motion_threshold, log,
                    width=args.md_width, dilates=args.md_dilates, blur=args.md_blur)

        if(audio_list is not None and motion_list is not None):
            if(len(audio_list) > len(motion_list)):
                audio_list = audio_list[:len(motion_list)]
            elif(len(motion_list) > len(audio_list)):
                motion_list = motion_list[:len(audio_list)]

        return combine_audio_motion(audio_list, motion_list, args.edit_based_on, log)

    has_loud = get_has_loud(inp, args, fps, audio_samples, sample_rate, log)
    has_loud_length = len(has_loud)
    has_loud = apply_mark_as(has_loud, has_loud_length, fps, args, log)
    has_loud = cook(has_loud, min_clip, min_cut)
    has_loud = apply_frame_margin(has_loud, has_loud_length, frame_margin)

    if(segment is not None):
        has_loud = combine_segment(has_loud, segment, fps)
    # Remove small clips/cuts created by applying other rules.
    has_loud = cook(has_loud, min_clip, min_cut)
    return apply_spacing_rules(has_loud, has_loud_length, min_clip, min_cut, speeds,
        fps, args, log)


def edit_media(i, inp, ffmpeg, args, progress, speeds, segment, exporting_to_editor,
    data_file, temp, log):
    from auto_editor.utils.container import get_rules

    chunks = None
    if(inp.ext == '.json'):
        from auto_editor.formats.make_json import read_json_cutlist

        input_path, chunks, speeds = read_json_cutlist(inp.path, auto_editor.version, log)
        inp = ffmpeg.file_info(input_path)

        output_path = set_output_name(inp.path, inp.ext, data_file, args)
    else:
        output_path = args.output_file[i]
        if(not os.path.isdir(inp.path) and '.' not in output_path):
            output_path = set_output_name(output_path, inp.ext, data_file, args)

    log.debug('{} -> {}'.format(inp.path, output_path))

    output_container = os.path.splitext(output_path)[1].replace('.', '')

    # Check if export options make sense.
    rules = get_rules(output_container)
    codec_error = "'{}' codec is not supported in '{}' container."

    if(not fnone(args.sample_rate)):
        if(rules['samplerate'] is not None and args.sample_rate not in rules['samplerate']):
            log.error("'{}' container only supports samplerates: {}".format(output_container,
                rules['samplerate']))

    vcodec = args.video_codec
    if(vcodec == 'uncompressed'):
        vcodec = 'mpeg4'
    if(vcodec == 'copy'):
        vcodec = inp.video_streams[0]['codec']

    if(vcodec != 'auto'):
        if(rules['vstrict'] and vcodec not in rules['vcodecs']):
            log.error(codec_error.format(vcodec, output_container))

        if(vcodec in rules['disallow_v']):
            log.error(codec_error.format(vcodec, output_container))

    acodec = args.audio_codec
    if(acodec == 'copy'):
        acodec = inp.audio_streams[0]['codec']
        log.debug(f'Settings acodec to {acodec}')

    if(acodec not in ['unset', 'auto']):
        if(rules['astrict'] and acodec not in rules['acodecs']):
            log.error(codec_error.format(acodec, output_container))

        if(acodec in rules['disallow_a']):
            log.error(codec_error.format(acodec, output_container))

    if(args.keep_tracks_seperate and rules['max_audio_streams'] == 1):
        log.warning(
            "'{}' container doesn't support multiple audio tracks.".format(container)
        )

    if(os.path.isfile(output_path) and inp.path != output_path and not args.preview):
        log.debug('Removing already existing file: {}'.format(output_path))
        os.remove(output_path)

    audio_samples = None
    tracks = len(inp.audio_streams)
    audio_file = len(inp.video_streams) == 0 and tracks == 1

    if(inp.fps is None):
        fps = 30
    elif(exporting_to_editor):
        fps = int(float(inp.fps))
    else:
        fps = float(inp.fps)

    if(fps < 1):
        log.error('{}: Frame rate cannot be below 1. fps: {}'.format(inp.basename, fps))

    # Extract subtitles in their native format.
    if(len(inp.subtitle_streams) > 0):
        cmd = ['-i', inp.path, '-hide_banner']
        for s, sub in enumerate(inp.subtitle_streams):
            cmd.extend(['-map', '0:s:{}'.format(s)])
        for s, sub in enumerate(inp.subtitle_streams):
            cmd.extend([os.path.join(temp, '{}s.{}'.format(s, sub['ext']))])
        ffmpeg.run(cmd)

    sample_rate = None

    if(args.cut_by_this_track >= tracks and 'cut_by_this_track' in args._set):
        message = "You choose a track that doesn't exist.\nThere "
        if(tracks == 1):
            message += 'is only {} track.\n'.format(tracks)
        else:
            message += 'are only {} tracks.\n'.format(tracks)
        for t in range(tracks):
            message += ' Track {}\n'.format(t)
        log.error(message)

    def number_of_VFR_frames(text, log):
        import re
        search = re.search(r'VFR:[\d.]+ \(\d+\/\d+\)', text, re.M)
        if(search is None):
            log.warning('Could not get number of VFR Frames.')
            return 0
        else:
            nums = re.search(r'\d+\/\d+', search.group()).group(0)
            log.debug('VFR Frames: {}'.format(nums))
            return int(nums.split('/')[0])

    def has_VFR(cmd, log):
        return number_of_VFR_frames(ffmpeg.pipe(cmd), log) != 0

    # Split audio tracks into: 0.wav, 1.wav, etc.
    cmd = ['-i', inp.path, '-hide_banner']
    for t in range(tracks):
        cmd.extend(['-map', '0:a:{}'.format(t), '-ac', '2',
            os.path.join(temp, '{}.wav'.format(t))])
    cmd.extend(['-map', '0:v:0'])
    if(args.has_vfr == 'unset' and len(inp.video_streams) > 0 and not args.preview):
        log.conwrite('Extracting audio / detecting VFR')
        cmd.extend(['-vf', 'vfrdet', '-f', 'null', '-'])
        has_vfr = has_VFR(cmd, log)
    else:
        log.conwrite('Extracting audio')
        ffmpeg.run(cmd)
        has_vfr = args.has_vfr == 'yes'
    del cmd

    if(len(inp.video_streams) > 0 and tracks == 0):
        # Doesn't matter because we don't need to align to an audio track.
        has_vfr = False

    log.debug('Has VFR: {}'.format(has_vfr))

    if(tracks != 0):
        if(args.cut_by_all_tracks):
            temp_file = os.path.join(temp, 'combined.wav')
            cmd = ['-i', inp.path, '-filter_complex',
                '[0:a]amix=inputs={}:duration=longest'.format(tracks), '-ac', '2',
                '-f', 'wav', temp_file]
            ffmpeg.run(cmd)
            del cmd
        else:
            temp_file = os.path.join(temp, '{}.wav'.format(args.cut_by_this_track))

        from auto_editor.scipy.wavfile import read
        sample_rate, audio_samples = read(temp_file)

    log.debug('Frame Rate: {}'.format(fps))
    if(chunks is None):
        chunks = get_chunks(inp, speeds, segment, fps, args, log, audio_samples,
            sample_rate)

    def is_clip(chunk):
        # type: (list) -> bool
        return speeds[chunk[2]] != 99999

    def number_of_cuts(chunks, speeds):
        # type: (list, list) -> int
        return len(list(filter(is_clip, chunks)))

    def get_clips(chunks, speeds):
        clips = []
        for chunk in chunks:
            if(is_clip(chunk)):
                clips.append([chunk[0], chunk[1], speeds[chunk[2]] * 100])
        return clips

    num_cuts = number_of_cuts(chunks, speeds)
    clips = get_clips(chunks, speeds)

    effects = Effect(args, log, _vars={})

    if(args.export_as_json):
        from auto_editor.formats.make_json import make_json_cutlist
        make_json_cutlist(inp.path, output_path, auto_editor.version, chunks, speeds,
            log)
        return num_cuts, output_path

    if(args.preview):
        from auto_editor.preview import preview
        preview(inp, chunks, speeds, log)
        return num_cuts, None

    if(args.export_to_premiere):
        from auto_editor.formats.premiere import premiere_xml
        premiere_xml(inp, temp, output_path, clips, chunks, sample_rate, audio_file,
            fps, log)
        return num_cuts, output_path

    if(args.export_to_final_cut_pro):
        from auto_editor.formats.final_cut_pro import fcp_xml

        total_frames = chunks[len(chunks) - 1][1]
        fcp_xml(inp, temp, output_path, clips, tracks, total_frames, audio_file, fps, log)
        return num_cuts, output_path

    if(args.export_to_shotcut):
        from auto_editor.formats.shotcut import shotcut_xml

        shotcut_xml(inp, temp, output_path, clips, chunks, fps, log)
        return num_cuts, output_path

    def pad_chunk(item, total_frames):
        start = None
        end = None
        if(item[0] != 0):
            start = [0, item[0], 2]
        if(item[1] != total_frames -1):
            end = [item[1], total_frames -1, 2]

        if(start is None):
            return [item] + [end]
        if(end is None):
            return [start] + [item]
        return [start] + [item] + [end]


    def make_media(inp, chunks, output_path):
        from auto_editor.utils.video import mux_quality_media

        if(rules['allow_subtitle']):
            from auto_editor.render.subtitle import cut_subtitles
            cut_subtitles(ffmpeg, inp, chunks, speeds, fps, temp, log)

        if(rules['allow_audio']):
            from auto_editor.render.audio import make_new_audio

            for t in range(tracks):
                temp_file = os.path.join(temp, '{}.wav'.format(t))
                new_file = os.path.join(temp, 'new{}.wav'.format(t))
                make_new_audio(temp_file, new_file, chunks, speeds, log, fps, progress)

                if(not os.path.isfile(new_file)):
                    log.bug('Audio file not created.')

        video_stuff = []

        if(rules['allow_video']):
            from auto_editor.render.video import render_av
            for v, vid in enumerate(inp.video_streams):
                if(vid['codec'] in ['png', 'jpeg']):
                    video_stuff.append(('image', None, None))
                else:
                    video_stuff.append(render_av(ffmpeg, v, inp, args, chunks, speeds,
                        fps, has_vfr, progress, effects, rules, temp, log))

        log.conwrite('Writing the output file.')

        mux_quality_media(ffmpeg, video_stuff, rules, output_path, output_container,
            args, inp, temp, log)
        if(output_path is not None and not os.path.isfile(output_path)):
            log.bug(f'The file {output_path} was not created.')

    total_frames = chunks[len(chunks) - 1][1]

    if(args.export_as_clip_sequence):
        speeds.append(99999)
        num_clips = 0
        for chunk in chunks:
            if(speeds[chunk[2]] == 99999):
                continue
            make_media(inp, pad_chunk(chunk, total_frames),
                append_filename(output_path, '-{}'.format(num_clips)))
            num_clips += 1
    else:
        make_media(inp, chunks, output_path)
    return num_cuts, output_path



'''utils/effects.py'''

from PIL import Image, ImageDraw, ImageFont

class Effect():
    __slots__ = ('all', 'sheet', '_vars', 'log', 'background', 'width', 'height')
    def __init__(self, args, log, _vars):
        self.all = []
        self.sheet = {}
        self._vars = _vars
        self.log = log
        self.background = args.background

    def add_var(self, key, item):
        self._vars[key] = item

    def _values(self, val, _type):
        # just to be clear, val can only be None if the default value is None.
        # Users can't set the argument to None themselves.
        if(val is None):
            return None

        if(_type is str):
            return str(val)

        for key, item in self._vars.items():
            if(val == key):
                return _type(item)

        if(not isinstance(val, int)
            and not (val.replace('.', '', 1)).replace('-', '', 1).isdigit()):
            self.log.error("variable '{}' is not defined.".format(val))

        return _type(val)

    def set_timing(self, effect, effect_index):
        start = effect.pop('start')
        dur = effect.pop('dur')

        start = self._values(start[1], int)
        dur = self._values(dur[1], int)

        if(dur < 1):
            self.log.error(f"dur's value must be greater than 0. Was '{dur}'.")

        for i in range(start, start + dur, 1):
            if(i in self.sheet):
                self.sheet[i].append(effect_index)
            else:
                self.sheet[i] = [effect_index]

        return effect

    def set_all(self, effect):
        this_type = effect['type']
        new_effect = {'type': this_type}

        for name, my_tup in effect.items():
            if(name != 'type'):
                _type, val = my_tup
                if(isinstance(val, str) and val.startswith('{') and val.endswith('}')):
                    new_effect[name] = new_effect[val[1:-1]]
                else:
                    new_effect[name] = self._values(val, _type)

        return new_effect

    def resolve(self, args):
        self.width = self._vars['width']
        self.height = self._vars['height']

        effect_list = args.add_text + args.add_rectangle + args.add_ellipse

        for i, my_effect in enumerate(effect_list):
            effect = self.set_timing(my_effect, i)
            effect = self.set_all(effect)

            # Change font attr to ImageFont Obj. (must have font and size attr)
            if('font' in effect):
                if(effect['font'] == 'default'):
                    try:
                        effect['font'] = ImageFont.truetype(effect['font'], effect['size'])
                    except OSError:
                        effect['font'] = ImageFont.load_default()
                else:
                    try:
                        effect['font'] = ImageFont.truetype(effect['font'], effect['size'])
                    except OSError:
                        self.log.error(f"Font '{effect['font']}' not found.")

            self.all.append(effect)

    def apply(self, index, frame, pix_fmt):
        img = frame.to_image().convert('RGBA')

        for item in self.sheet[index]:
            pars = self.all[item]

            obj_img = Image.new('RGBA', img.size, (255,255,255,0))
            draw = ImageDraw.Draw(obj_img)

            if(pars['type'] == 'text'):

                tw, th = draw.textsize(pars['content'], font=pars['font'])

                new_x = ((pars['x'] * 2) - tw) / 2
                new_y = ((pars['y'] * 2) - th) / 2

                draw.text((new_x, new_y), pars['content'], font=pars['font'],
                    fill=pars['fill'])

            if(pars['type'] == 'rectangle'):
                draw.rectangle([pars['x1'], pars['y1'], pars['x2'], pars['y2']],
                    fill=pars['fill'], width=pars['width'], outline=pars['outline'])

            if(pars['type'] == 'ellipse'):
                draw.ellipse([pars['x1'], pars['y1'], pars['x2'], pars['y2']],
                    fill=pars['fill'], width=pars['width'], outline=pars['outline'])

            img = Image.alpha_composite(img, obj_img)

        frame = frame.from_image(img).reformat(format=pix_fmt)
        return frame



'''io/wav.py'''

import wave
import numpy as np

class WavReader():
    def __init__(self, filename):
        self._reader = wave.open(filename, 'rb')

    @property
    def channels(self):
        return self._reader.getnchannels()

    @property
    def empty(self):
        return self._reader.tell() == self._reader.getnframes()

    def close(self):
        self._reader.close()

    def read(self, buffer):
        if(buffer.shape[0] != self.channels):
            raise ValueError("the buffer should have the same number of "
                "channels as the WavReader")

        frames = self._reader.readframes(buffer.shape[1])
        frames = np.frombuffer(frames, '<i2').astype(np.float32) / 32676

        # Separate channels
        frames = frames.reshape((-1, self.channels)).T

        n = frames.shape[1]
        np.copyto(buffer[:, :n], frames)
        del frames

        return n

    @property
    def samplerate(self):
        return self._reader.getframerate()

    @property
    def samplewidth(self):
        return self._reader.getsamplewidth()

    def skip(self, n):
        current_pos = self._reader.tell()
        new_pos = min(current_pos + n, self._reader.getnframes())

        self._reader.setpos(new_pos)

        return new_pos - current_pos

    def __enter__(self):
        return self

    def __exit__(self, _1, _2, _3):
        self.close()


class WavWriter():
    def __init__(self, filename, channels, samplerate):
        self._writer = wave.open(filename, 'wb')
        self._channels = channels
        self._writer.setnchannels(channels)
        self._writer.setframerate(samplerate)
        self._writer.setsampwidth(2)

    @property
    def channels(self):
        return self._channels

    def close(self):
        self._writer.close()

    def write(self, buffer):
        if(buffer.shape[0] != self.channels):
            raise ValueError(f"buffer has {buffer.shape[0]} channels while self has {self.channels} channels")

        np.clip(buffer, -1, 1, out=buffer)

        n = buffer.shape[1]
        frames = (buffer.T.reshape((-1,)) * 32676).astype(np.int16).tobytes()
        self._writer.writeframes(frames)
        del frames

        return n

    def __enter__(self):
        return self

    def __exit__(self, _1, _2, _3):
        self.close()



'''analyze/generic.py'''

import math
import numpy as np

def get_np_list(inp, audio_samples, sample_rate, fps, func):
    if(audio_samples is not None):
        sample_count = audio_samples.shape[0]
        sample_rate_per_frame = sample_rate / fps
        audio_frame_count = int(math.ceil(sample_count / sample_rate_per_frame))
        return func((audio_frame_count), dtype=np.bool_)

    import cv2
    cap = cv2.VideoCapture(inp.path)
    total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT)) + 1
    return func((total_frames), dtype=np.bool_)



'''setup.py'''

import re
from setuptools import setup, find_packages

def pip_version():
    with open('auto_editor/__init__.py') as f:
        version_content = f.read()

    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
        version_content, re.M)

    if(version_match):
        return version_match.group(1)

    raise ValueError('Unable to find version string.')

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='auto-editor',
    version=pip_version(),
    description='Auto-Editor: Effort free video editing!',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='Unlicense',
    url='https://auto-editor.com',
    project_urls={
        'Bug Tracker': 'https://github.com/WyattBlue/auto-editor/issues',
        'Source Code': 'https://github.com/WyattBlue/auto-editor',
    },
    author='WyattBlue',
    author_email='wyattblue@auto-editor.com',
    keywords='video audio media editor editing processing nonlinear automatic '
     'silence-detect silence-removal silence-speedup motion-detection',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'numpy>=1.19.5',
        'yt-dlp>=2021.11.10.1',
        'pillow>=5.3.0',
        'av>=8.0.3',
        'opencv-python>=4.3',
    ],
    python_requires='>=3.7,<3.12',
    classifiers=[
        'Topic :: Multimedia :: Sound/Audio',
        'Topic :: Multimedia :: Video',
        'License :: Public Domain',
        'License :: OSI Approved :: The Unlicense (Unlicense)',
        'Environment :: Console',
        'Natural Language :: English',
        'Intended Audience :: End Users/Desktop',
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    entry_points={
        "console_scripts": [
            "auto-editor=auto_editor.__main__:main",
            "aecreate=auto_editor.subcommands.create:main",
            "aedesc=auto_editor.subcommands.desc:main",
            "aeinfo=auto_editor.subcommands.info:main",
            "aesubdump=auto_editor.subcommands.subdump:main",
            "aegrep=auto_editor.subcommands.grep:main",
        ]
    }
)



'''vanparse.py'''

import sys
import difflib

class ParserError(Exception):
    pass

def indent(text, prefix, predicate=None):
    if predicate is None:
        def predicate(line):
            return line.strip()

    def prefixed_lines():
        for line in text.splitlines(True):
            yield (prefix + line if predicate(line) else line)
    return ''.join(prefixed_lines())

def out(text):
    import re
    import textwrap

    from shutil import get_terminal_size
    width = get_terminal_size().columns - 3

    indent_regex = re.compile(r'^(\s+)')
    wrapped_lines = []

    for line in text.split('\n'):
        exist_indent = re.search(indent_regex, line)
        pre_indent = exist_indent.groups()[0] if exist_indent else ''

        wrapped_lines.append(
            textwrap.fill(line, width=width, subsequent_indent=pre_indent)
        )

    print('\n'.join(wrapped_lines))

def print_option_help(args, option):
    text = '  ' + ', '.join(option['names']) + '\n    ' + option['help'] + '\n\n'
    if(option['keywords'] != []):
        text += '    Arguments:\n    '
        for n, item in enumerate(option['keywords']):
            [[k, v]] = item.items()

            if(k != 'type'):
                _, default = v
                text += '{' + k
                text += '' if default == '' else '=' + str(default)
                text += '}' if n + 1 == len(option['keywords']) else '},'

    if(option['manual'] != ''):
        text += '{}\n\n'.format(indent(option['manual'], '    '))

    if(option['keywords'] != []):
        pass
    elif(option['action'] == 'default'):
        text += '    type: ' + option['type'].__name__
        text += '\n    default: {}\n'.format(option['default'])
        if(option['range'] is not None):
            text += '    range: ' +  option['range'] + '\n'

        if(option['choices'] is not None):
            text += '    choices: ' +  ', '.join(option['choices']) + '\n'
    elif(option['action'] in ['store_true', 'store_false']):
        text += '    type: flag\n'
    else:
        text += '    type: unknown\n'

    out(text)

def print_program_help(the_args):
    text = ''
    for options in the_args:
        for option in options:
            if(option['action'] == 'text'):
                text += '\n  ' + option['names'][0] + '\n'
            elif(option['action'] == 'blank'):
                text += '\n'
            elif(not option['hidden']):
                text += '  ' + ', '.join(option['names']) + ': ' + option['help'] + '\n'
    text += '\n'
    out(text)

def get_option(item, the_args):
    for options in the_args:
        for option in options:
            if(option['action'] != 'text' and option['action'] != 'blank'):
                dash = list(map(lambda n: n.replace('_', '-'), option['names']))
                if((item in option['names'] or item in dash)):
                    return option
    return None

def _to_key(val):
    # (val: dict) -> str
    return val['names'][0].replace('-', '')

class ArgumentParser():
    def __init__(self, program_name, version, description=None):
        self.program_name = program_name
        self._version = version
        self.description = description

        self.args = []
        self.kwarg_defaults = {
            'nargs': 1,
            'type': str,
            'default': None,
            'action': 'default',
            'range': None,
            'choices': None,
            'group': None,
            'help': '',
            'keywords': [],
            'hidden': False,
            'manual': '',
        }

    def add_argument(self, *args, **kwargs):
        my_dict = {
            'names': list(args),
        }

        for key, item in self.kwarg_defaults.items():
            my_dict[key] = item

        for key, item in kwargs.items():
            if(key not in self.kwarg_defaults):
                raise ValueError('key {} not found.'.format(key))
            my_dict[key] = item

        self.args.append(my_dict)

    def add_text(self, text):
        self.args.append({
            'names': [text],
            'action': 'text',
        });

    def add_blank(self):
        self.args.append({'action': 'blank'});

    def parse_args(self, sys_args):
        if(sys_args == [] and self.description):
            out(self.description)
            sys.exit()

        if(sys_args == ['-v'] or sys_args == ['-V']):
            out('{} version {}'.format(self.program_name, self._version))
            sys.exit()

        return ParseOptions(sys_args, self.args)


"""
    Positional Arguments
        --rectangle 0,end,10,20,20,30,#000, ...

    Keyword Arguments
        --rectangle start=0,end=end,x1=10, ...

"""

class ParseOptions():

    @staticmethod
    def parse_parameters(unsplit_arguments, op):

        ARG_SEP = ','
        KEYWORD_SEP = '='
        REQUIRED = ''

        dic = {}
        keys = []
        for key in op['keywords']:
            [[k, v]] = key.items()
            keys.append(k)
            dic[k] = v

        allow_positional_args = True

        for i, item in enumerate(unsplit_arguments.split(ARG_SEP)):
            if(i+1 > len(keys)):
                raise ParserError(f"Too many arguments, starting with '{item}'.")

            if(KEYWORD_SEP in item):
                allow_positional_args = False

                parameters = item.split(KEYWORD_SEP)
                if(len(parameters) > 2):
                    raise ParserError(f"Invalid Syntax: '{item}'.")
                key, val = parameters
                if(key not in dic):
                    raise ParserError(f"Parameter '{key}' does not exist.")
                dic[key] = (dic[key][0], val)
            elif(allow_positional_args):
                dic[keys[i]] = (dic[keys[i]][0], item)
            else:
                raise ParserError('Positional argument follows keyword argument.')

        # Check if any required args are not specified.
        for key, type_val in dic.items():
            if(type_val[1] == REQUIRED):
                raise ParserError(f"Parameter '{key}' is required.")
        return dic

    def __init__(self, sys_args, *args):
        # Set the default options.
        for options in args:
            for option in options:
                if(option['action'] == 'text' or option['action'] == 'blank'):
                    continue

                key = _to_key(option)
                if(option['action'] == 'store_true'):
                    value = False
                elif(option['action'] == 'store_false'):
                    value = True
                elif(option['nargs'] != 1):
                    value = []
                else:
                    value = option['default']
                setattr(self, key, value)

        # Figure out command line options changed by user.
        my_list = []
        used_options = []
        _set = []
        setting_inputs = True
        option_list = 'input'
        list_type = str
        i = 0
        while i < len(sys_args):
            item = sys_args[i]
            label = 'option' if item.startswith('--') else 'short'

            option = get_option(item, the_args=args)

            def error_message(args, item, label):

                def all_names(args):
                    name_set = set()
                    for options in args:
                        for opt in options:
                            if(opt['action'] not in ['text', 'blank']):
                                for names in opt['names']:
                                    name_set.add(names)
                    return name_set

                opt_list = all_names(args)
                close_matches = difflib.get_close_matches(item, opt_list)
                if(close_matches):
                    return 'Unknown {}: {}\n\n    Did you mean:\n        '.format(
                        label, item) + ', '.join(close_matches)
                return 'Unknown {}: {}'.format(label, item)

            if(option is None):
                # Unknown Option!
                if(setting_inputs and (option_list != 'input' or (option_list == 'input' and not item.startswith('-')))):
                    # Option is actually an input file, like example.mp4

                    if(option_list != 'input'):
                        _op = used_options[-1]
                        if(_op['keywords'] != []):
                            item = self.parse_parameters(item, _op)
                    my_list.append(item)
                else:
                    raise ParserError(error_message(args, item, label))
            else:
                # We found the option.
                if(option_list is not None):
                    setattr(self, option_list, list(map(list_type, my_list)))

                setting_inputs = False
                option_list = None
                my_list = []

                if(option in used_options):
                    raise ParserError('Cannot repeat option {} twice.'.format(option['names'][0]))

                used_options.append(option)

                key = _to_key(option)
                _set.append(key)

                nextItem = None if i == len(sys_args) - 1 else sys_args[i+1]
                if(nextItem == '-h' or nextItem == '--help'):
                    print_option_help(args, option)
                    sys.exit()

                if(option['nargs'] != 1):
                    setting_inputs = True
                    option_list = key
                    list_type = option['type']
                elif(option['action'] == 'store_true'):
                    value = True
                elif(option['action'] == 'store_false'):
                    value = False
                else:
                    value = option['type'](nextItem)

                    if(option['choices'] is not None and value not in option['choices']):
                        option_name = option['names'][0]
                        my_choices = ', '.join(option['choices'])

                        raise ParserError('{} is not a choice for {}\nchoices are:\n  {}'.format(
                            value, option_name, my_choices))
                    i += 1
                setattr(self, key, value)

            i += 1
        if(setting_inputs):
            setattr(self, option_list, list(map(list_type, my_list)))
        setattr(self, '_set', _set)
        if(self.help):
            print_program_help(args)
            sys.exit()



'''subcommands/levels.py'''

import sys

def levels_options(parser):
    parser.add_argument('--kind', type=str, default='audio')
    parser.add_argument('--track', type=int, default=0,
        help='what audio/video track to get. If --kind is set to motion, track will look at video streams.')
    parser.add_argument('--ffmpeg_location', default=None,
        help='point to your custom ffmpeg file.')
    parser.add_argument('--my_ffmpeg', action='store_true',
        help='use the ffmpeg on your PATH instead of the one packaged.')
    parser.add_argument('--help', '-h', action='store_true',
        help='print info about the program or an option and exit.')
    parser.add_argument('input', nargs='*',
        help='the template')
    return parser

def main(sys_args=sys.argv[1:]):
    import os
    import tempfile

    import auto_editor
    import auto_editor.vanparse as vanparse

    from auto_editor.utils.log import Log
    from auto_editor.ffwrapper import FFmpeg

    parser = vanparse.ArgumentParser('levels', auto_editor.version,
        description='Get loudness of audio over time.')
    parser = levels_options(parser)

    temp = tempfile.mkdtemp()
    log = Log(temp=temp)

    try:
        args = parser.parse_args(sys_args)
    except vanparse.ParserError as e:
        log.error(str(e))

    ffmpeg = FFmpeg(args.ffmpeg_location, args.my_ffmpeg, False)

    inp = ffmpeg.file_info(args.input[0])
    fps = 30 if inp.fps is None else float(inp.fps)

    if(args.kind == 'audio'):
        from auto_editor.analyze.audio import display_audio_levels

        if(args.track >= len(inp.audio_streams)):
            log.error("Audio track '{}' does not exist.".format(args.track))

        read_track = os.path.join(temp, '{}.wav'.format(args.track))

        ffmpeg.run(['-i', inp.path, '-ac', '2', '-map', '0:a:{}'.format(args.track),
            read_track])

        if(not os.path.isfile(read_track)):
            log.error('Audio track file not found!')

        display_audio_levels(read_track, fps)

    if(args.kind == 'motion'):
        if(args.track >= len(inp.video_streams)):
            log.error("Video track '{}' does not exist.".format(args.track))

        from auto_editor.analyze.motion import display_motion_levels

        display_motion_levels(inp, width=400, dilates=2, blur=21)

    log.cleanup()

if(__name__ == '__main__'):
    main()



'''utils/log.py'''

import sys
from shutil import rmtree, get_terminal_size
from time import perf_counter, sleep
from datetime import timedelta

class Timer():
    __slots__ = ['start_time', 'quiet']
    def __init__(self, quiet=False):
        self.start_time = perf_counter()
        self.quiet = quiet

    def stop(self):
        if(not self.quiet):
            second_len = round(perf_counter() - self.start_time, 2)
            minute_len = timedelta(seconds=round(second_len))

            sys.stdout.write(f'Finished. took {second_len} seconds ({minute_len})\n')

class Log():
    __slots__ = ['is_debug', 'quiet', 'temp']
    def __init__(self, show_debug=False, quiet=False, temp=None):
        self.is_debug = show_debug
        self.quiet = quiet
        self.temp = temp

    def debug(self, message):
        if(self.is_debug):
            self.conwrite('')
            sys.stderr.write(f'Debug: {message}\n')

    def cleanup(self):
        if(self.temp is None):
            return
        try:
            rmtree(self.temp)
            self.debug('Removed Temp Directory.')
        except PermissionError:
            sleep(0.1)
            try:
                rmtree(self.temp)
                self.debug('Removed Temp Directory.')
            except Exception:
                self.debug('Failed to delete temp dir.')
        except FileNotFoundError:
            # that's ok, the folder we are trying to remove is already gone
            pass

    def conwrite(self, message):
        if(not self.quiet):
            buffer = get_terminal_size().columns - len(message) - 3
            sys.stdout.write('  ' + message + ' ' * buffer + '\r')
            try:
                sys.stdout.flush()
            except AttributeError:
                pass

    def error(self, message):
        self.conwrite('')
        message = message.replace('\t', '    ')
        sys.stderr.write(f'Error! {message}\n')
        self.cleanup()

        from platform import system

        if(system() == 'Linux'):
            sys.exit(1)
        else:
            try:
                sys.exit(1)
            except SystemExit:
                import os
                os._exit(1)

    def bug(self, message, bug_type='bug report'):
        self.conwrite('')
        URL = 'https://github.com/WyattBlue/auto-editor/issues/'

        sys.stderr.write(
            'Error! {}\n\nSomething went wrong!\nCreate a {} at:\n  {}\n'.format(
                message, bug_type, URL))
        self.cleanup()
        sys.exit(1)

    def warning(self, message):
        if(not self.quiet):
            sys.stderr.write(f'Warning! {message}\n')

    def print(self, message):
        if(not self.quiet):
            sys.stdout.write(f'{message}\n')

    def checkType(self, data, name, correct_type):
        if(not isinstance(data, correct_type)):
            badtype = type(data).__name__
            goodtype = correct_type.__name__
            self.bug('Variable "{}" was not a {}, but a {}'.format(
                name, goodtype, badtype), 'bug report')
